//////////////////////////////////////////////////////////////////
//
// ProxyThread.cxx
//
// Copyright (c) Citron Network Inc. 2001-2002
//
// This work is published under the GNU Public License (GPL)
// see file COPYING for details.
// We also explicitely grant the right to link this code
// with the OpenH323 library.
//
// initial author: Chih-Wei Huang <cwhuang@linux.org.tw>
// initial version: 12/7/2001
//
//////////////////////////////////////////////////////////////////

#if (_MSC_VER >= 1200)
#pragma warning( disable : 4355 ) // warning about using 'this' in initializer
#pragma warning( disable : 4786 ) // warning about too long debug symbol off
#pragma warning( disable : 4800 ) // warning about forcing value to bool
#endif

#include <algorithm>
#include <functional>
#include "ANSI.h"
#include "gk.h"
#include "gk_const.h"
#include "stl_supp.h"
#include "ProxyThread.h"
#include "rwlock.h"


class ProxyConnectThread : public MyPThread {
public:
	PCLASSINFO ( ProxyConnectThread, MyPThread )

	ProxyConnectThread(ProxyHandleThread *);
	bool Acquire();
	void Release() { available = true; }
	bool Connect(ProxySocket *);

	// override from class MyPThread
	virtual void Exec();

private:
	ProxyHandleThread *handler;
	bool available;

	ProxySocket *calling;
	PMutex amutex;
};

struct TPKTV3 {
	TPKTV3() {}
	TPKTV3(WORD);

	BYTE header, padding;
	WORD length;
};

inline TPKTV3::TPKTV3(WORD len)
	: header(3), padding(0)
{
	length = PIPSocket::Host2Net(WORD(len + sizeof(TPKTV3)));
}


// class ProxySocket
ProxySocket::ProxySocket(PIPSocket *s, const char *t) : self(s), type(t)
{
	wbufsize = 1536; // 1.5KB
	wbuffer = new BYTE[wbufsize];
	buflen = 0;
	blocked = connected = deletable = false;
}

ProxySocket::~ProxySocket()
{
	delete [] wbuffer;
	std::for_each(queue.begin(), queue.end(), delete_array);
	PTRACE(4, type << "\tDelete socket " << name);
}

ProxySocket::Result ProxySocket::ReceiveData()
{
	if (!self->Read(wbuffer, wbufsize)) {
		ErrorHandler(PChannel::LastReadError);
		return NoData;
	}
	PTRACE(6, type << "\tReading from " << Name());
	buflen = self->GetLastReadCount();
	return Forwarding;
}

bool ProxySocket::ForwardData()
{
	return WriteData(wbuffer, buflen);
}

bool ProxySocket::TransmitData(const PBYTEArray & buf)
{
	return WriteData(buf, buf.GetSize());
}

bool ProxySocket::TransmitData(const BYTE *buf, WORD len)
{
	return WriteData(buf, len);
}

bool ProxySocket::EndSession()
{
	MarkBlocked(false);
//	SetConnected(false);
	return CloseSocket();
}

bool ProxySocket::Flush()
{
	while (queue.size() > 0) {
		queueMutex.Wait();
		PBYTEArray *pdata = queue.front();
		queue.pop_front();
		queueMutex.Signal();
		PBYTEArray data(*pdata);
		delete pdata;
		if (!WriteData(data, data.GetSize()))
			return false;
		PTRACE(4, type << '\t' << data.GetSize() << " be flushed to " << Name());
	}
	MarkBlocked(false);
	return true;
}

bool ProxySocket::WriteData(const BYTE *buf, WORD len)
{
	if (!IsSocketOpen())
		return false;

	if (writeMutex.WillBlock()) {
		PTRACE(3, Name() << " socket is busy, " << len << " bytes queued");
		PWaitAndSignal qlock(queueMutex);
		queue.push_back(new PBYTEArray(buf, len));
		return false;
	}

	PWaitAndSignal lock(writeMutex);

	MarkBlocked(true);
	if (self->Write(buf, len)) {
		MarkBlocked(false);
		PTRACE(6, Name() << ' ' << len << " bytes sent");
		return true;
	}

	WORD rcount = self->GetLastWriteCount();
	buf += rcount, len -= rcount;

	PTRACE(4, Name() << " blocked, " << rcount << " bytes written, " << len << " bytes queued");
	queueMutex.Wait();
	queue.push_front(new PBYTEArray(buf, len));
	queueMutex.Signal();
	return ErrorHandler(PChannel::LastWriteError);
}

bool ProxySocket::ErrorHandler(PChannel::ErrorGroup group)
{
	PChannel::Errors e = self->GetErrorCode(group);

	PString msg(PString(type) + "\t" + Name());
	switch (e)
	{
	//	case PChannel::NoError:
	//	// I don't know why there is error with code NoError
	//		PTRACE(4, msg << " Error(" << group << "): No error?");
	//		break;
		case PChannel::Timeout:
			PTRACE(4, msg << " Error(" << group << "): Timeout");
			break;
		default:
			PTRACE(3, msg << " Error(" << group << "): " << PChannel::GetErrorText(e) << " (" << e << ')');
			MarkBlocked(false);
			CloseSocket();
			break;
	}
	return false;
}

void ProxySocket::SetName(PIPSocket::Address ip, WORD pt)
{
	name = ip.AsString() + PString(PString::Printf, ":%u", pt);
}

// class TCPProxySocket
TCPProxySocket::TCPProxySocket(const char *t, TCPProxySocket *s, WORD p)
      : PTCPSocket(p), ProxySocket(this, t), remote(s)
{
}

TCPProxySocket::~TCPProxySocket()
{
	if (remote) {
		remote->remote = 0; // detach myself from remote
		remote->SetDeletable();
	}
}

bool TCPProxySocket::ForwardData()
{
	return (remote) ? remote->InternalWrite(buffer) : false;
}

bool TCPProxySocket::TransmitData(const PBYTEArray & buf)
{
	return InternalWrite(buf);
}

BOOL TCPProxySocket::Accept(PSocket & socket)
{
//	SetReadTimeout(PMaxTimeInterval);
	BOOL result = PTCPSocket::Accept(socket);
	PTimeInterval timeout(100);
	SetReadTimeout(timeout);
	SetWriteTimeout(timeout);
	// since GetName() may not work if socket closed,
	// we save it for reference
	Address ip;
	WORD pt;
	GetPeerAddress(ip, pt);
	SetName(ip, pt);
	return result;
}

BOOL TCPProxySocket::Connect(WORD localPort, const Address & addr)
{
	SetName(addr, GetPort());
	SetReadTimeout(PTimeInterval(6000)); // TODO: read from config...
	BOOL result = PTCPSocket::Connect(localPort, addr);
	PTimeInterval timeout(100);
	SetReadTimeout(timeout);
	SetWriteTimeout(timeout);
	return result;
}

bool TCPProxySocket::ReadTPKT()
{
	PTRACE(5, Type() << "\tReading from " << Name());
	if (buflen == 0) {
		TPKTV3 tpkt;
		if (!ReadBlock(&tpkt, sizeof(TPKTV3)))
			return ErrorHandler(LastReadError);
		//if (tpkt.header != 3 || tpkt.padding != 0)
		// some bad endpoints don't set padding to 0, e.g., Cisco AS5300
		if (tpkt.header != 3) {
			PTRACE(2, "Proxy\t" << Name() << " NOT TPKT PACKET!");
			return false; // Only support version 3
		}
		buflen = Net2Host(tpkt.length) - sizeof(TPKTV3);
		if (buflen < 1) {
			PTRACE(3, "Proxy\t" << Name() << " PACKET TOO SHORT!");
			buflen = 0;
			return false;
		}
		if (!SetMinBufSize(buflen))
			return false;
		buffer = PBYTEArray(bufptr = wbuffer, buflen, false);
	}

	if (!Read(bufptr, buflen))
		return ErrorHandler(LastReadError);

	if (!(buflen -= GetLastReadCount()))
		return true;

	bufptr += GetLastReadCount();
	PTRACE(3, "Proxy\t" << Name() << " read timeout?");
	return false;
}

bool TCPProxySocket::InternalWrite(const PBYTEArray & buf)
{
	WORD len = buf.GetSize(), tlen = len + sizeof(TPKTV3);
	PBYTEArray tbuf(tlen);
	BYTE *bptr = tbuf.GetPointer();
	new (bptr) TPKTV3(len); // placement operator
	memcpy(bptr + sizeof(TPKTV3), buf, len);
	return WriteData(bptr, tlen);
}

bool TCPProxySocket::SetMinBufSize(WORD len)
{
	if (wbufsize < len) {
		delete [] wbuffer;
		wbuffer = new BYTE[wbufsize = len];
	}
	return (wbuffer != 0);
}


// class MyPThread
MyPThread::MyPThread() : PThread(5000, NoAutoDeleteThread), isOpen(true)
{
}

void MyPThread::Close()
{
	isOpen = false;
	sync.Signal();
}

bool MyPThread::Wait()
{
	ReadUnlock unlock(ConfigReloadMutex);
	sync.Wait();
	return isOpen;
}

void MyPThread::Go()
{
	if (sync.WillBlock())
		sync.Signal();
}

void MyPThread::Main()
{
	PTRACE(2, GetClass() << ' ' << getpid() << " started");
	while (isOpen)
		Exec();

	PTRACE(2, GetClass() << ' ' << getpid() << " closed");
}

bool MyPThread::Destroy()
{
	Close();
	WaitForTermination();
	delete this;
	return true; // useless, workaround for VC
}


// class ProxyConnectThread
ProxyConnectThread::ProxyConnectThread(ProxyHandleThread *h)
	: handler(h), available(true)
{
	Resume();
}

bool ProxyConnectThread::Acquire()
{
	PWaitAndSignal lock(amutex);
	if (!available)
		return false;
	available = false;
	return true;
}

bool ProxyConnectThread::Connect(ProxySocket *socket)
{
	if (available) {
		PTRACE(1, "ProxyC\tWarning: not acquired yet");
		return false;
	}

	socket->MarkBlocked(true);
	calling = socket;
	Go();
	return true;
}

void ProxyConnectThread::Exec()
{
	if (!Wait())
		return;

	TCPProxySocket *socket = dynamic_cast<TCPProxySocket *>(calling);
	TCPProxySocket *remote = socket->ConnectTo();
	if (remote) {
		handler->Insert(remote);
		socket->MarkBlocked(false);
	}
	// else
	// 	Note: socket may be invalid

	available = true;
}

// class ProxyListener
ProxyListener::ProxyListener(HandlerList *h, PIPSocket::Address i, WORD p, unsigned qs)
      : m_listener(0), m_interface(i), m_port(p), m_handler(h)
{
	Open(qs);
}

ProxyListener::~ProxyListener()
{
	delete m_listener;
}

bool ProxyListener::Open(unsigned queueSize)
{
	m_listener = new PTCPSocket(m_port);
	isOpen = m_listener->Listen(m_interface, queueSize, m_port, PSocket::CanReuseAddress);
	m_port = m_listener->GetPort(); // get the listen port
	if (isOpen) {
		PTRACE(2, "ProxyL\tListen to " << m_interface << ':' << m_port);
		Resume();
#ifdef PTRACING
	} else {
		PTRACE(1, "ProxyL\tCan't listen port " << m_port);
#endif
	}
	return isOpen;
}

void ProxyListener::Close()
{
	PTRACE(3, "ProxyL\tClosing ProxyListener");
	m_listener->Close();
	isOpen = false;
}

void ProxyListener::Exec()
{
	if (!m_listener->IsOpen()) {
		isOpen = false;
		return;
	}

	TCPProxySocket *socket = CreateSocket();
	if (socket->Accept(*m_listener)) { // incoming connection
		PTRACE(3, "ProxyL\tConnected from " << socket->Name());
		m_handler->Insert(socket);
	} else {
		PChannel::Errors err = socket->GetErrorCode();
		delete socket;  // delete unused socket
		PTRACE(3, "ProxyL\tError: " << PChannel::GetErrorText(err));
	}
}

// class ProxyHandleThread
ProxyHandleThread::ProxyHandleThread(PINDEX i)
{
	SetID(PString(PString::Printf, "ProxyH(%u)", i));
	Resume();

	lcHandler = new ProxyHandleThread;
	lcHandler->SetID(PString(PString::Printf, "ProxyLC(%u)", i));
	lcHandler->SetPriority(HighPriority);
	lcHandler->Resume();
}

ProxyHandleThread::~ProxyHandleThread()
{
	std::for_each(connList.begin(), connList.end(), mem_fun(&MyPThread::Destroy));
	std::for_each(sockList.begin(), sockList.end(), delete_socket);
	// The RTP/RTCP sockets should be deleted after
	// call signalling sockets being deleted
	if (lcHandler)
		lcHandler->Destroy();
}

void ProxyHandleThread::Insert(ProxySocket *socket)
{
	socket->SetHandler(this);
	mutex.StartWrite();
	sockList.push_back(socket);
	mutex.EndWrite();
	Go();
}

void ProxyHandleThread::FlushSockets()
{
	PSocket::SelectList wlist;
	mutex.StartRead();
	iterator i = sockList.begin(), j = sockList.end();
	while (i != j) {
		if ((*i)->CanFlush())
			(*i)->AddToSelectList(wlist);
		++i;
	}
	mutex.EndRead();
	if (wlist.IsEmpty())
		return;

	// unfortunately, there is no method to select only sockets to write
	PSocket::SelectList rlist;
	if (PSocket::Select(rlist, wlist, PTimeInterval(10)) != PChannel::NoError)
	       return;
	
	PTRACE(5, "Proxy\t" << wlist.GetSize() << " sockets to flush...");
	for (PINDEX k = 0; k < wlist.GetSize(); ++k) {
		ProxySocket *socket = dynamic_cast<ProxySocket *>(&wlist[k]);
		if (socket->Flush()) {
			PTRACE(4, "Proxy\t" << socket->Name() << " flush ok");
		}
	}
}

void ProxyHandleThread::Remove(iterator i)
{
	PWaitAndSignal lock(removedMutex);
	removedList.push_back(*i);
	sockList.erase(i);
}

void ProxyHandleThread::Remove(ProxySocket *socket)
{
	PWaitAndSignal lock(removedMutex);
	removedList.push_back(socket);
	sockList.remove(socket);
}

void ProxyHandleThread::RemoveSockets()
{
	std::for_each(removedList.begin(), removedList.end(), delete_socket);
	PWaitAndSignal lock(removedMutex);
	removedList.clear();
}

void ProxyHandleThread::BuildSelectList(PSocket::SelectList & result)
{
	WriteLock lock(mutex);
	iterator i = sockList.begin(), j = sockList.end();
	while (i != j) {
		iterator k=i++;
		ProxySocket *socket = *k;
		if (!socket->IsBlocked()) {
			if (socket->IsSocketOpen())
				socket->AddToSelectList(result);
			else if (!socket->IsConnected()) {
				Remove(k);
				continue;
			}
			if (socket->IsDeletable())
				Remove(k);
#ifdef PTRACING
	//	} else {
	//		PTRACE(5, socket->Name() << " is busy!");
#endif
		}
	}
}

void ProxyHandleThread::Exec()
{
	PINDEX ss;
	ReadLock cfglock(ConfigReloadMutex);
	PSocket::SelectList sList;
	while (true) {
		FlushSockets();
		BuildSelectList(sList);
		if ((ss = sList.GetSize()))
			break;
		RemoveSockets();
		PTRACE(5, id << " waiting...");
		if (!Wait())
			return;
	}

	if (ss) {
		ReadUnlock cfglock(ConfigReloadMutex);
		PChannel::Errors r = PSocket::Select(sList, PTimeInterval(100));
		if (r != PChannel::NoError || sList.IsEmpty()) {
			PTRACE_IF(3, r !=PChannel::NoError, id << " Select error: " << r);
			RemoveSockets();
			return;
		}
	}

#ifdef PTRACING
	PString msg(PString::Printf, " %u sockets selected from %u, total %u/%u", sList.GetSize(), ss, sockList.size(), removedList.size());
	PTRACE(4, id + msg);
#endif
	ss = sList.GetSize();
	for (PINDEX i = 0; i < ss; ++i) {
		ProxySocket *socket = dynamic_cast<ProxySocket *>(&sList[i]);
		switch (socket->ReceiveData())
		{
			case ProxySocket::Connecting:
				ConnectTo(socket);
				break;
			case ProxySocket::Forwarding:
				if (!socket->ForwardData()) {
					PTRACE(3, "Proxy\t" << socket->Name() << " forward blocked");
				}
				break;
			case ProxySocket::Closing:
				socket->ForwardData();
				// then close the socket
			case ProxySocket::Error:
				socket->CloseSocket();
				break;
			default:
				break;
		}
	}
	RemoveSockets();
}

ProxyConnectThread *ProxyHandleThread::FindConnectThread()
{
	connMutex.StartRead();
	for (citerator i = connList.begin(); i != connList.end(); ++i)
		if ((*i)->Acquire()) {
			connMutex.EndRead();
			return (*i);
		}

	connMutex.EndRead();
	ProxyConnectThread *ct = new ProxyConnectThread(this);
	connMutex.StartWrite();
	connList.push_back(ct);
	connMutex.EndWrite();
	PTRACE(2, "Proxy\tCreate a new ConnectThread, total " << connList.size());
	ct->Acquire();
	return ct;
}

void ProxyHandleThread::ConnectTo(ProxySocket *socket)
{
	FindConnectThread()->Connect(socket);
}

bool ProxyHandleThread::CloseUnusedThreads()
{
	static unsigned idx = 0;
	ProxyConnectThread *ct = 0;
	if (connList.size() <= 1 || (++idx % 60))
		return true;

	connMutex.StartRead();
	citerator i = connList.end();
	do {
		if ((*--i)->Acquire()) {
			if (!ct) {
				ct = (*i);
				continue;
			}
			connMutex.EndRead();
			connMutex.StartWrite();
			connList.remove(ct);
			connMutex.EndWrite();
			(*i)->Release();
			ct->Destroy();
			PTRACE(1, id << " Close one unused ConnectThread, left " << connList.size());
			return true;
		}
	} while (i != connList.begin());
	connMutex.EndRead();
	if (ct)
		ct->Release();
	return true; // useless, workaround for VC
}


HandlerList::HandlerList(PIPSocket::Address home) : GKHome(home), GKPort(0)
{
	currentHandler = 0;
	listenerThread = 0;
	LoadConfig();
}

HandlerList::~HandlerList()
{
	CloseListener();
	std::for_each(handlers.begin(), handlers.end(), std::mem_fun(&MyPThread::Destroy));
}

void HandlerList::Insert(ProxySocket *socket)
{
	// only lister thread will call this method, no mutex required
	handlers[currentHandler]->Insert(socket);
	if (++currentHandler >= handlers.size())
		currentHandler = 0;
}

void HandlerList::Check()
{
	std::for_each(handlers.begin(), handlers.end(), std::mem_fun(&ProxyHandleThread::CloseUnusedThreads));
}

void HandlerList::CloseListener()
{
	if (listenerThread) {
		listenerThread->Destroy();
		listenerThread = 0;
	}
}

